from . import utils
from .table import TableFinder
from .container import Container
from copy import copy

from pdfminer.pdftypes import resolve_all
from six import string_types
import re
lt_pat = re.compile(r"^LT")

class Page(Container):
    cached_properties = Container.cached_properties + [ "_layout" ]
    is_original = True

    def __init__(self, pdf, page_obj, page_number=None, initial_doctop=0):
        self.pdf = pdf
        self.page_obj = page_obj
        self.page_number = page_number
        _rotation = self.decimalize(resolve_all(self.page_obj.attrs.get("Rotate", 0)))
        self.rotation =  _rotation % 360
        self.page_obj.rotate = self.rotation
        self.initial_doctop = self.decimalize(initial_doctop)

        cropbox = page_obj.attrs.get("CropBox")
        mediabox = page_obj.attrs.get("MediaBox")

        self.cropbox = self.decimalize(resolve_all(cropbox)) if cropbox is not None else None
        self.mediabox = self.decimalize(resolve_all(mediabox) or self.cropbox)
        m = self.mediabox

        if self.rotation in [ 90, 270 ]:
            self.bbox = self.decimalize((
                min(m[1], m[3]),
                min(m[0], m[2]),
                max(m[1], m[3]),
                max(m[0], m[2]),
            ))
        else:
            self.bbox = self.decimalize((
                min(m[0], m[2]),
                min(m[1], m[3]),
                max(m[0], m[2]),
                max(m[1], m[3]),
            ))

    def decimalize(self, x):
        return utils.decimalize(x, self.pdf.precision)

    @property
    def width(self):
        return self.bbox[2] - self.bbox[0]

    @property
    def height(self):
        return self.bbox[3] - self.bbox[1]

    @property
    def layout(self):
        if hasattr(self, "_layout"): return self._layout
        self._layout = self.pdf.process_page(self.page_obj)
        return self._layout

    @property
    def objects(self):
        if hasattr(self, "_objects"): return self._objects
        self._objects = self.parse_objects()
        return self._objects

    def parse_objects(self):
        objects = {}

        d = self.decimalize
        h = self.height
        idc = self.initial_doctop
        pno = self.page_number

        def point2coord(pt):
            x, y = pt
            return (
                d(x),
                h - d(y)
            )

        noop = lambda x: x
        str_conv = lambda x: str(x or "")

        CONVERSIONS = {
            # Decimals
            "adv": d,
            "height": d,
            "linewidth": d,
            "pts": d,
            "size": d,
            "srcsize": d,
            "width": d,
            "x0": d,
            "x1": d,
            "y0": d,
            "y1": d,

            # Integer
            "bits": int,
            "upright": int,

            # Strings
            "font": str_conv,
            "fontname": str_conv,
            "imagemask": noop,
            "name": str_conv,
            "object_type": str_conv,
            "text": str_conv,

            # No conversion
            "colorspace": noop,
            "evenodd": noop,
            "fill": noop,
            "non_stroking_color": noop,
            "path": noop,
            "stream": noop,
            "stroke": noop,
            "stroking_color": noop,
        }

        CONVERSIONS_KEYS = set(CONVERSIONS.keys())

        def process_object(obj):
            attr = dict((k, CONVERSIONS[k](resolve_all(v)))
                for k, v in obj.__dict__.items()
                    if k in CONVERSIONS_KEYS)

            kind = re.sub(lt_pat, "", obj.__class__.__name__).lower()
            attr["object_type"] = kind
            attr["page_number"] = pno

            if hasattr(obj, "get_text"):
                attr["text"] = obj.get_text()

            if kind == "curve":
                attr["points"] = list(map(point2coord, obj.pts))

            if attr.get("y0") != None:
                attr["top"] = h - attr["y1"]
                attr["bottom"] = h - attr["y0"]
                attr["doctop"] = idc + attr["top"]

            if objects.get(kind) == None:
                objects[kind] = []
            objects[kind].append(attr)

            if hasattr(obj, "_objs"):
                for child in obj._objs:
                    process_object(child)
        
        for obj in self.layout._objs:
            process_object(obj)

        return objects

    def debug_tablefinder(self, table_settings={}):
        return TableFinder(self, table_settings)

    def find_tables(self, table_settings={}):
        return TableFinder(self, table_settings).tables

    def extract_tables(self, table_settings={}):
        tables = self.find_tables(table_settings)

        extract_kwargs = dict((k, table_settings["text_" + k]) for k in [
            "x_tolerance",
            "y_tolerance",
        ] if "text_" + k in table_settings)

        return [ table.extract(**extract_kwargs) for table in tables ]

    def extract_table(self, table_settings={}):
        tables = self.find_tables(table_settings)
        # Return the largest table, as measured by number of cells.
        sorter = lambda x: (-len(x.cells), x.bbox[1], x.bbox[0])
        largest = list(sorted(tables, key=sorter))[0]
        return largest.extract()

    def extract_text(self,
        x_tolerance=utils.DEFAULT_X_TOLERANCE,
        y_tolerance=utils.DEFAULT_Y_TOLERANCE):

        return utils.extract_text(self.chars,
            x_tolerance=x_tolerance,
            y_tolerance=y_tolerance)

    def extract_words(self,
        x_tolerance=utils.DEFAULT_X_TOLERANCE,
        y_tolerance=utils.DEFAULT_Y_TOLERANCE,
        keep_blank_chars=False):

        return utils.extract_words(self.chars,
            x_tolerance=x_tolerance,
            y_tolerance=y_tolerance,
            keep_blank_chars=keep_blank_chars)

    def crop(self, bbox):
        class CroppedPage(DerivedPage):
            @property
            def objects(self):
                if hasattr(self, "_objects"): return self._objects
                self._objects = utils.crop_to_bbox(
                    self.parent_page.objects,
                    self.bbox)
                return self._objects

        cropped = CroppedPage(self)
        cropped.bbox = self.decimalize(bbox)
        return cropped

    def within_bbox(self, bbox):
        """
        Same as .crop, except only includes objects fully within the bbox
        """
        class CroppedPage(DerivedPage):
            @property
            def objects(self):
                if hasattr(self, "_objects"): return self._objects
                self._objects = utils.within_bbox(
                    self.parent_page.objects,
                    self.bbox)
                return self._objects

        cropped = CroppedPage(self)
        cropped.bbox = self.decimalize(bbox)
        return cropped

    def filter(self, test_function):
        class FilteredPage(DerivedPage):
            @property
            def objects(self):
                if hasattr(self, "_objects"): return self._objects
                self._objects = utils.filter_objects(
                    self.parent_page.objects,
                    test_function
                )
                return self._objects
        filtered = FilteredPage(self)
        filtered.bbox = self.bbox
        return filtered

    def to_image(self, **conversion_kwargs):
        """
        For conversion_kwargs, see http://docs.wand-py.org/en/latest/wand/image.html#wand.image.Image
        """
        from .display import PageImage, DEFAULT_RESOLUTION
        kwargs = dict(conversion_kwargs)
        if "resolution" not in conversion_kwargs:
            kwargs["resolution"] = DEFAULT_RESOLUTION
        return PageImage(self, **kwargs)

class DerivedPage(Page):
    is_original = False
    def __init__(self, parent_page):
        self.parent_page = parent_page
        self.pdf = parent_page.pdf
        self.page_obj = parent_page.page_obj
        self.page_number = parent_page.page_number
        self.flush_cache(Container.cached_properties)

        if type(parent_page) == Page:
            self.root_page = parent_page
        else:
            self.root_page = parent_page.root_page



#!/usr/bin/env python
import pdfplumber
import argparse
from itertools import chain
try:
    from cdecimal import Decimal, ROUND_HALF_UP
except ImportError:
    from decimal import Decimal, ROUND_HALF_UP
import unicodecsv
import codecs
import json
import sys

class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return float(o.quantize(Decimal('.0001'), rounding=ROUND_HALF_UP))
        return super(DecimalEncoder, self).default(o)

def parse_page_spec(p_str):
    if "-" in p_str:
        return list(range(*map(int, p_str.split("-"))))
    else: return [ int(p_str) ]

def parse_args():
    parser = argparse.ArgumentParser("pdfplumber")

    stdin = (sys.stdin.buffer if sys.version_info[0] >= 3 else sys.stdin)
    parser.add_argument("infile", nargs="?",
        type=argparse.FileType("rb"),
        default=stdin)

    parser.add_argument("--format",
        choices=["csv", "json"],
        default="csv")

    parser.add_argument("--encoding",
        default="utf-8")

    TYPE_DEFAULTS = [ "char", "anno", "line", "curve", "rect" ]
    parser.add_argument("--types", nargs="+",
        choices=TYPE_DEFAULTS + [ "rect_edge" ],
        default=TYPE_DEFAULTS)

    parser.add_argument("--pages", nargs="+",
        type=parse_page_spec)

    args = parser.parse_args()
    if args.pages != None:
        args.pages = list(chain(*args.pages))
    return args

def to_csv(pdf, types, encoding):
    objs = []
    fields = set()
    for t in types:
        new_objs = getattr(pdf, t + "s")
        if len(new_objs):
            objs += new_objs
            fields = fields.union(set(new_objs[0].keys()))

    first_columns = [
        "object_type", "page_number",
        "x0", "x1", "y0", "y1",
        "doctop", "top", "bottom",
        "width", "height"
    ]

    cols = first_columns + list(sorted(set(fields) - set(first_columns)))
    stdout = (sys.stdout.buffer if sys.version_info[0] >= 3 else sys.stdout)
    w = unicodecsv.DictWriter(stdout,
        fieldnames=cols, encoding=encoding)
    w.writeheader()
    w.writerows(objs)

def to_json(pdf, types, encoding):
    data = { "metadata": pdf.metadata }

    def get_page_data(page):
        d = dict((t + "s", getattr(page, t + "s"))
            for t in types)
        d["width"] = page.width
        d["height"] = page.height
        return d

    data["pages"] = list(map(get_page_data, pdf.pages))

    if hasattr(sys.stdout, "buffer"):
        sys.stdout = codecs.getwriter("utf-8")(sys.stdout.buffer, "strict")
        json.dump(data, sys.stdout, cls=DecimalEncoder)
    else:
        json.dump(data, sys.stdout, cls=DecimalEncoder, encoding=encoding)

def main():
    args = parse_args()
    pdf = pdfplumber.load(args.infile, pages=args.pages)
    if args.format == "csv":
        to_csv(pdf, args.types, args.encoding)
    else:
        to_json(pdf, args.types, args.encoding)

if __name__ == "__main__":
    main()



from pdfminer.converter import PDFLayoutAnalyzer
from pdfminer.utils import mult_matrix, apply_matrix_pt, get_bound
from pdfminer.layout import LTFigure
import six

def add_wh(attr):
    attr.update({
        "height": attr["y1"] - attr["y0"],
        "width": attr["x1"] - attr["x0"],
    })
    return attr

def render_image(self, name, stream):
    srcsize = (stream.get_any(("W", "Width")),
        stream.get_any(("H", "Height")))
    imagemask = stream.get_any(("IM", "ImageMask"))
    bits = stream.get_any(("BPC", "BitsPerComponent"), 1)
    colorspace = stream.get_any(("CS", "ColorSpace"))
    if not isinstance(colorspace, list):
        colorspace = [colorspace]

    item = {
        "name": name,
        "srcsize": srcsize,
        "imagemask": imagemask,
        "colorspace": colorspace,
        "bits": bits,
        "stream": stream,
        "x0": self.cur_item.x0,
        "y0": self.cur_item.y0,
        "x1": self.cur_item.x1,
        "y1": self.cur_item.y1,
        "object_type": "image",
    }
    self.cur_item.add(add_wh(item))
    return

PDFLayoutAnalyzer.render_image = render_image

def paint_path(self, gstate, stroke, fill, evenodd, path):
    shape = ''.join(x[0] for x in path)
    if shape == 'ml':
        # horizontal/vertical line
        (_, x0, y0) = path[0]
        (_, x1, y1) = path[1]
        (x0, y0) = apply_matrix_pt(self.ctm, (x0, y0))
        (x1, y1) = apply_matrix_pt(self.ctm, (x1, y1))
        if x0 == x1 or y0 == y1:
            x0, y0, x1, y1 = get_bound([ (x0, y0), (x1, y1) ])

            item = add_wh({
                "x0": x0,
                "y0": y0,
                "x1": x1,
                "y1": y1,
                "object_type": "line",
            })

            if self.interpreter.parse_styles:
                item.update({
                    "stroke_width": 0 if stroke is False else gstate.linewidth,
                    "stroke": None if stroke is False else (gstate.scolor_hex or "#000000"),
                    "fill": None if fill is False else (gstate.ncolor_hex or "#000000"),
                    "dash": gstate.dash,
                    "evenodd": evenodd,
                })

            self.cur_item.add(item)
            return

    if shape == 'mlllh':
        # rectangle
        (_, x0, y0) = path[0]
        (_, x1, y1) = path[1]
        (_, x2, y2) = path[2]
        (_, x3, y3) = path[3]
        (x0, y0) = apply_matrix_pt(self.ctm, (x0, y0))
        (x1, y1) = apply_matrix_pt(self.ctm, (x1, y1))
        (x2, y2) = apply_matrix_pt(self.ctm, (x2, y2))
        (x3, y3) = apply_matrix_pt(self.ctm, (x3, y3))
        if ((x0 == x1 and y1 == y2 and x2 == x3 and y3 == y0) or
            (y0 == y1 and x1 == x2 and y2 == y3 and x3 == x0)):
            y_min = min(y0, y1, y2, y3)
            y_max = max(y0, y1, y2, y3)
            x_min = min(x0, x1, x2, x3)
            x_max = max(x0, x1, x2, x3)
            item = add_wh({
                "x0": x_min,
                "y0": y_min,
                "x1": x_max,
                "y1": y_max,
                "object_type": "rect",
            })
            if self.interpreter.parse_styles:
                item.update({
                    "stroke_width": 0 if stroke is False else gstate.linewidth,
                    "stroke": None if stroke is False else (gstate.scolor_hex or "#000000"),
                    "fill": None if fill is False else (gstate.ncolor_hex or "#000000"),
                    "dash": gstate.dash,
                    "evenodd": evenodd,
                })
            self.cur_item.add(item)
            return
    # other shapes
    pts = []
    for p in path:
        for i in range(1, len(p), 2):
            pts.append(apply_matrix_pt(self.ctm, (p[i], p[i+1])))
    x0, y0, x1, y1 = get_bound(pts)

    item = add_wh({
        "x0": x0,
        "y0": y0,
        "x1": x1,
        "y1": y1,
        "points": pts,
        "path": path,
        "object_type": "curve"
    })

    if self.interpreter.parse_styles:
        item.update({
            "stroke_width": 0 if stroke is False else gstate.linewidth,
            "stroke": None if stroke is False else (gstate.scolor_hex or "#000000"),
            "fill": None if fill is False else (gstate.ncolor_hex or "#000000"),
            "dash": gstate.dash,
            "evenodd": evenodd,
        })

    self.cur_item.add(item)
    return

PDFLayoutAnalyzer.paint_path = paint_path

def render_char(self, matrix, font, fontsize, scaling, rise, cid):
    try:
        text = font.to_unichr(cid)
        assert isinstance(text, six.text_type), text
    except PDFUnicodeNotDefined:
        text = self.handle_undefined_char(font, cid)

    textwidth = font.char_width(cid)
    textdisp = font.char_disp(cid)
    adv = textwidth * fontsize * scaling

    # compute the boundary rectangle.
    if font.is_vertical():
        # vertical
        width = font.get_width() * fontsize
        (vx, vy) = textdisp
        if vx is None:
            vx = width * 0.5
        else:
            vx = vx * fontsize * .001
        vy = (1000 - vy) * fontsize * .001
        tx = -vx
        ty = vy + rise
        bll = (tx, ty+self.adv)
        bur = (tx+width, ty)
    else:
        # horizontal
        height = font.get_height() * fontsize
        descent = font.get_descent() * fontsize
        ty = descent + rise
        bll = (0, ty)
        bur = (adv, ty+height)
    (a, b, c, d, e, f) = matrix
    upright = (0 < a*d*scaling and b*c <= 0)
    (x0, y0) = apply_matrix_pt(matrix, bll)
    (x1, y1) = apply_matrix_pt(matrix, bur)
    if x1 < x0:
        (x0, x1) = (x1, x0)
    if y1 < y0:
        (y0, y1) = (y1, y0)

    if font.is_vertical():
        size = x1 - x0
    else:
        size = y1 - y0

    # cf. "Table 106 – Text rendering modes" of PDF 1.7 spec
    gs = self.interpreter.graphicstate
    r = self.interpreter.textstate.render
    if r in (0, 2, 4, 6):
        fill = gs.ncolor_hex
    else:
        fill = None
    if r in (1, 2, 5, 6):
        stroke = gs.scolor_hex
    else:
        stroke = None

    item = add_wh({
        "x0": x0,
        "y0": y0,
        "x1": x1,
        "y1": y1,
        "fontname": font.fontname,
        "size": size,
        "text": text,
        "object_type": "char",
    })

    if self.interpreter.parse_styles:
        item.update({
            "fill": fill,
            "stroke": stroke,
        })

    self.cur_item.add(item)
    return adv

PDFLayoutAnalyzer.render_char = render_char



from pdfminer.pdfinterp import (
    PDFGraphicState,
    PDFPageInterpreter,
    PDFInterpreterError,
)

from pdfminer.pdfcolor import (
    LITERAL_DEVICE_GRAY,
    LITERAL_DEVICE_RGB,
    LITERAL_DEVICE_CMYK,
)

from ..utils import decode_psl_list

import spectra

def pdfcolor_to_hex(c, cs_name):
    if c is None:
        # Per PDF 1.7 spec, default is black
        return "#000000"
    if isinstance(c, (tuple, list)):
        c = decode_psl_list(c) 
    else:
        c = decode_psl_list([ c ])[0] 
        if type(c) is str and c[0] == "P":
            c = c[1:]

    if cs_name in (None, "DeviceGray"):
        # Per PDF 1.7 spec, default is DeviceGray
        s = spectra.rgb(c, c, c)
    elif cs_name == "DeviceRGB":
        s = spectra.rgb(*c)
    elif cs_name == "DeviceCMYK":
        s = spectra.cmyk(*c)
    else:
        raise ValueError("Cannot handle '{}' color space. Try parsing wth parse_colors=False".format(cs_name))

    return s.hexcode

# PDFGraphicState
def __init__(self):
    self.linewidth = 0
    self.linecap = None
    self.linejoin = None
    self.miterlimit = None
    self.dash = None
    self.intent = None
    self.flatness = None

    # stroking color
    self.scolor = None
    self.scs = None

    # non stroking color
    self.ncolor = None
    self.ncs = None
    return

def copy(self):
    obj = PDFGraphicState()
    obj.linewidth = self.linewidth
    obj.linecap = self.linecap
    obj.linejoin = self.linejoin
    obj.miterlimit = self.miterlimit
    obj.dash = self.dash
    obj.intent = self.intent
    obj.flatness = self.flatness
    obj.scolor = self.scolor
    obj.scs = self.scs
    obj.ncolor = self.ncolor
    obj.ncs = self.ncs
    return obj

@property
def scolor_hex(self):
    cs_name = "DeviceGray" if self.scs is None else self.scs.name
    return pdfcolor_to_hex(self.scolor, cs_name)

@property
def ncolor_hex(self):
    cs_name = "DeviceGray" if self.ncs is None else self.ncs.name 
    return pdfcolor_to_hex(self.ncolor, cs_name)
    
PDFGraphicState.__init__ = __init__
PDFGraphicState.copy = copy
PDFGraphicState.scolor_hex = scolor_hex
PDFGraphicState.ncolor_hex = ncolor_hex

# PDFPageInterpreter

# setgray-stroking
def do_G(self, gray):
    self.graphicstate.scolor = gray
    self.do_CS(LITERAL_DEVICE_GRAY)
    return

# setgray-non-stroking
def do_g(self, gray):
    self.graphicstate.ncolor = gray
    self.do_cs(LITERAL_DEVICE_GRAY)
    return

# setrgb-stroking
def do_RG(self, r, g, b):
    self.graphicstate.scolor = (r, g, b)
    self.do_CS(LITERAL_DEVICE_RGB)
    return

# setrgb-non-stroking
def do_rg(self, r, g, b):
    self.graphicstate.ncolor = (r, g, b)
    self.do_cs(LITERAL_DEVICE_RGB)
    return

# setcmyk-stroking
def do_K(self, c, m, y, k):
    self.graphicstate.scolor = (c, m, y, k)
    self.do_CS(LITERAL_DEVICE_CMYK)
    return

# setcmyk-non-stroking
def do_k(self, c, m, y, k):
    self.graphicstate.ncolor = (c, m, y, k)
    self.do_cs(LITERAL_DEVICE_CMYK)
    return

# setcmyk-non-stroking
def do_k(self, c, m, y, k):
    self.graphicstate.ncolor = (c, m, y, k)
    self.do_cs(LITERAL_DEVICE_CMYK)
    return

# setcolor
def do_SCN(self):
    if self.scs:
        n = self.scs.ncomponents
    else:
        if settings.STRICT:
            raise PDFInterpreterError('No colorspace specified!')
        n = 1
    self.graphicstate.scolor = self.pop(n)
    return

def do_scn(self):
    if self.ncs:
        n = self.ncs.ncomponents
    else:
        if settings.STRICT:
            raise PDFInterpreterError('No colorspace specified!')
        n = 1
    self.graphicstate.ncolor = self.pop(n)
    return

def __init__(self, rsrcmgr, device, parse_styles=False):
    self.rsrcmgr = rsrcmgr
    self.device = device
    self.device.interpreter = self
    self.parse_styles = parse_styles
#
#    if parse_styles:
#        self.do_G = do_G
#        self.do_g = do_g
#        self.do_RG = do_RG
#        self.do_rg = do_rg
#        self.do_K = do_K
#        self.do_k = do_k
#        self.do_SCN = do_SCN
#        self.do_scn = do_scn

PDFPageInterpreter.__init__ = __init__



from itertools import chain
from pdfplumber import utils

class Container(object):
    cached_properties = [ "_rect_edges", "_edges", "_objects" ]

    def flush_cache(self, properties=None):
        props = self.cached_properties if properties == None else properties
        for p in props:
            if hasattr(self, p):
                delattr(self, p)

    @property
    def rects(self):
        return self.objects.get("rect", [])

    @property
    def lines(self):
        return self.objects.get("line", [])

    @property
    def curves(self):
        return self.objects.get("curve", [])

    @property
    def images(self):
        return self.objects.get("image", [])

    @property
    def figures(self):
        return self.objects.get("figure", [])

    @property
    def chars(self):
        return self.objects.get("char", [])

    @property
    def annos(self):
        return self.objects.get("anno", [])

    @property
    def rect_edges(self):
        if hasattr(self, "_rect_edges"): return self._rect_edges
        rect_edges_gen = (utils.rect_to_edges(r) for r in self.rects)
        self._rect_edges = list(chain(*rect_edges_gen))
        return self._rect_edges

    @property
    def edges(self):
        if hasattr(self, "_edges"): return self._edges
        line_edges = list(map(utils.line_to_edge, self.lines))
        self._edges = self.rect_edges + line_edges
        return self._edges

    @property
    def horizontal_edges(self):
        test = lambda x: x["orientation"] == "h"
        return list(filter(test, self.edges))

    @property
    def vertical_edges(self):
        test = lambda x: x["orientation"] == "v"
        return list(filter(test, self.edges))



from pdfminer.utils import PDFDocEncoding
from pdfminer.psparser import PSLiteral
try:
    from cdecimal import Decimal, ROUND_HALF_UP
except ImportError:
    from decimal import Decimal, ROUND_HALF_UP
import numbers
from operator import itemgetter
import itertools
import six

if six.PY3:
    from functools import lru_cache as cache
else:
    # Python 2 has no lru_cache, so defining as a no-op
    def cache(**kwargs):
        def decorator(fn):
            return fn
        return decorator

DEFAULT_X_TOLERANCE = 3
DEFAULT_Y_TOLERANCE = 3

def cluster_list(xs, tolerance=0):
    tolerance = decimalize(tolerance)
    if tolerance == 0: return [ [x] for x in sorted(xs) ]
    if len(xs) < 2: return [ [x] for x in sorted(xs) ]
    groups = []
    xs = list(sorted(xs))
    current_group = [xs[0]]
    last = xs[0]
    for x in xs[1:]:
        if x <= (last + tolerance):
            current_group.append(x)
        else:
            groups.append(current_group)
            current_group = [x]
        last = x
    groups.append(current_group)
    return groups

def make_cluster_dict(values, tolerance):
    tolerance = decimalize(tolerance)
    clusters = cluster_list(set(values), tolerance)

    nested_tuples = [ [ (val, i) for val in value_cluster ]
        for i, value_cluster in enumerate(clusters) ]

    cluster_dict = dict(itertools.chain(*nested_tuples))
    return cluster_dict

def cluster_objects(objs, attr, tolerance):
    if isinstance(attr, (str, int)):
        attr_getter = itemgetter(attr)
    else:
        attr_getter = attr
    objs = to_list(objs)
    values = map(attr_getter, objs)
    cluster_dict = make_cluster_dict(values, tolerance)

    get_0, get_1 = itemgetter(0), itemgetter(1)

    cluster_tuples = sorted(((obj, cluster_dict.get(attr_getter(obj)))
        for obj in objs), key=get_1)

    grouped = itertools.groupby(cluster_tuples, key=get_1)

    clusters = [ list(map(get_0, v))
        for k, v in grouped ]

    return clusters

def decode_text(s):
    """
    Decodes a PDFDocEncoding string to Unicode.
    Adds py3 compatibility to pdfminer's version.
    """
    if type(s) == bytes and s.startswith(b'\xfe\xff'):
        return six.text_type(s[2:], 'utf-16be', 'ignore')
    else:
        ords = (ord(c) if type(c) == str else c for c in s)
        return ''.join(PDFDocEncoding[o] for o in ords)

def decode_psl_list(_list):
    return [ decode_text(value.name) if isinstance(value, PSLiteral) else value
        for value in _list ]

@cache(maxsize = int(10e4))
def _decimalize(v, q = None):
    # If already a decimal, just return itself
    if type(v) == Decimal:
        return v

    # If tuple/list passed, bulk-convert
    elif isinstance(v, (tuple, list)):
        return type(v)(decimalize(x, q) for x in v)

    # Convert int-like
    elif isinstance(v, numbers.Integral):
        return Decimal(int(v))

    # Convert float-like
    elif isinstance(v, numbers.Real):
        if q != None:
            return Decimal(repr(v)).quantize(Decimal(repr(q)),
                rounding=ROUND_HALF_UP)
        else:
            return Decimal(repr(v))
    else:
        raise ValueError("Cannot convert {0} to Decimal.".format(v))

def decimalize(v, q = None):
    # If already a decimal, just return itself
    if type(v) == Decimal:
        return v

    # If tuple/list passed, bulk-convert
    if isinstance(v, (tuple, list)):
        return type(v)(decimalize(x, q) for x in v)
    else:
        return _decimalize(v, q)

def is_dataframe(collection):
    cls = collection.__class__
    name = ".".join([ cls.__module__, cls.__name__ ])
    return name == "pandas.core.frame.DataFrame"

def to_list(collection):
    if is_dataframe(collection):
        return collection.to_dict("records")
    else:
        return list(collection)

def collate_line(line_chars, tolerance=DEFAULT_X_TOLERANCE):
    tolerance = decimalize(tolerance)
    coll = ""
    last_x1 = None
    for char in sorted(line_chars, key=itemgetter("x0")):
        if (last_x1 != None) and (char["x0"] > (last_x1 + tolerance)):
            coll += " "
        last_x1 = char["x1"]
        coll += char["text"]
    return coll

def objects_to_rect(objects):
    return {
        "x0": min(map(itemgetter("x0"), objects)),
        "x1": max(map(itemgetter("x1"), objects)),
        "top": min(map(itemgetter("top"), objects)),
        "bottom": max(map(itemgetter("bottom"), objects)),
    }

def objects_to_bbox(objects):
    return (
        min(map(itemgetter("x0"), objects)),
        min(map(itemgetter("top"), objects)),
        max(map(itemgetter("x1"), objects)),
        max(map(itemgetter("bottom"), objects)),
    )

obj_to_bbox = itemgetter("x0", "top", "x1", "bottom")

def bbox_to_rect(bbox):
    return {
        "x0": bbox[0],
        "top": bbox[1],
        "x1": bbox[2],
        "bottom": bbox[3]
    }



def extract_words(chars,
    x_tolerance=DEFAULT_X_TOLERANCE,
    y_tolerance=DEFAULT_Y_TOLERANCE,
    keep_blank_chars=False,
    horizontal_ltr = True, # Should words be read left-to-right?
    vertical_ttb = True, # Should vertical words be read top-to-bottom?
    ):

    x_tolerance = decimalize(x_tolerance)
    y_tolerance = decimalize(y_tolerance)

    def process_word_chars(chars, upright):
        x0, top, x1, bottom = objects_to_bbox(chars)

        if upright:
            if horizontal_ltr:
                sorted_chars = chars
            else:
                sorted_chars = sorted(chars, key = lambda x: -x["x1"])
        else:
            if vertical_ttb:
                sorted_chars = sorted(chars, key = itemgetter("doctop"))
            else:
                sorted_chars = sorted(chars, key = lambda x: -x["bottom"])

        return {
            "x0": x0,
            "x1": x1,
            "top": top,
            "bottom": bottom,
            "upright": upright,
            "text": "".join(map(itemgetter("text"), sorted_chars))
        }

    def get_line_words(chars, upright, tolerance=DEFAULT_X_TOLERANCE):
        get_text = itemgetter("text")
        min_key = "x0" if upright else "top"
        max_key = "x1" if upright else "bottom"

        chars_sorted = sorted(chars, key=itemgetter(min_key))

        words = []
        current_word = []

        for char in chars_sorted:
            if not keep_blank_chars and get_text(char).isspace():
                if len(current_word) > 0:
                    words.append(current_word)
                    current_word = []
                else: pass
            elif len(current_word) == 0:
                current_word.append(char)
            else:
                last_char = current_word[-1]
                if char[min_key] > (last_char[max_key] + tolerance):
                    words.append(current_word)
                    current_word = []
                current_word.append(char)

        if len(current_word) > 0:
            words.append(current_word)

        return [ process_word_chars(chars, upright)
            for chars in words ]

    chars_by_upright = { 1: [], 0: [] }
    words = []
    for char in to_list(chars):
        chars_by_upright[char.get("upright", 1)].append(char)

    for upright, char_group in chars_by_upright.items():
        clusters = cluster_objects(
            char_group,
            "doctop" if upright else "x0",
            y_tolerance, # Still use y-tolerance here, even for vertical words
        )

        for line_chars in clusters:
            words += get_line_words(line_chars, upright, tolerance = x_tolerance)

    return words

def extract_text(chars,
    x_tolerance=DEFAULT_X_TOLERANCE,
    y_tolerance=DEFAULT_Y_TOLERANCE):

    if len(chars) == 0:
        return None

    chars = to_list(chars)
    doctop_clusters = cluster_objects(chars, "doctop", y_tolerance)

    lines = (collate_line(line_chars, x_tolerance)
        for line_chars in doctop_clusters)

    coll = "\n".join(lines)
    return coll

collate_chars = extract_text

def filter_objects(objs, fn):
    if isinstance(objs, dict):
        return dict((k, filter_objects(v, fn))
            for k,v in objs.items())

    initial_type = type(objs)
    objs = to_list(objs)
    filtered = filter(fn, objs)

    return initial_type(filtered)

def get_bbox_overlap(a, b):
    a_left, a_top, a_right, a_bottom = decimalize(a)
    b_left, b_top, b_right, b_bottom = decimalize(b)
    o_left = max(a_left, b_left)
    o_right = min(a_right, b_right)
    o_bottom = min(a_bottom, b_bottom)
    o_top = max(a_top, b_top)
    o_width = o_right - o_left
    o_height = o_bottom - o_top
    if o_height >= 0 and o_width >= 0 and o_height + o_width > 0:
        return (o_left, o_top, o_right, o_bottom)
    else:
        return None

def clip_obj(obj, bbox):
    bbox = decimalize(bbox)

    overlap = get_bbox_overlap(obj_to_bbox(obj), bbox)
    if overlap is None: return None

    dims = bbox_to_rect(overlap)
    copy = dict(obj)

    for attr in [ "x0", "top", "x1", "bottom" ]:
        copy[attr] = dims[attr]

    if dims["top"] != obj["bottom"] or dims["top"] != obj["bottom"]:
        diff = dims["top"] - obj["top"]
        copy["doctop"] = obj["doctop"] + diff

    copy["width"] = copy["x1"] - copy["x0"]
    copy["height"] = copy["bottom"] - copy["top"]

    return copy

def intersects_bbox(objs, bbox):
    """
    Filters objs to only those intersecting the bbox
    """
    initial_type = type(objs)
    objs = to_list(objs)
    matching = [ obj for obj in objs
        if get_bbox_overlap(obj_to_bbox(obj), bbox) is not None ]
    return initial_type(matching)

def within_bbox(objs, bbox):
    """
    Filters objs to only those fully within the bbox
    """
    if isinstance(objs, dict):
        return dict((k, within_bbox(v, bbox))
            for k,v in objs.items())

    initial_type = type(objs)
    objs = to_list(objs)
    matching = [ obj for obj in objs
        if get_bbox_overlap(obj_to_bbox(obj), bbox) == obj_to_bbox(obj) ]
    return initial_type(matching)

def crop_to_bbox(objs, bbox):
    """
    Filters objs to only those intersecting the bbox,
    and crops the extent of the objects to the bbox.
    """
    if isinstance(objs, dict):
        return dict((k, crop_to_bbox(v, bbox))
            for k,v in objs.items())

    initial_type = type(objs)
    objs = to_list(objs)
    cropped = list(filter(None, (clip_obj(obj, bbox) for obj in objs)))
    return initial_type(cropped)

def move_object(obj, axis, value):
    assert(axis in ("h", "v"))
    if axis == "h":
        new_items = (
            ("x0", obj["x0"] + value),
            ("x1", obj["x1"] + value),
        )
    if axis == "v":
        new_items = [
            ("top", obj["top"] + value),
            ("bottom", obj["bottom"] + value),
        ]
        if "doctop" in obj:
            new_items += [ ("doctop", obj["doctop"] + value) ]
        if "y0" in obj:
            new_items += [
                ("y0", obj["y0"] - value),
                ("y1", obj["y1"] - value),
            ]
    return obj.__class__(tuple(obj.items()) + tuple(new_items))

def resize_object(obj, key, value):
    assert(key in ("x0", "x1", "top", "bottom"))
    old_value = obj[key]
    diff = value - old_value
    if key in ("x0", "x1"):
        if key == "x0":
            assert(value <= obj["x1"])
        else:
            assert(value >= obj["x0"])
        new_items = (
            (key, value),
            ("width", obj["width"] + diff),
        )
    if key == "top":
        assert(value <= obj["bottom"])
        new_items = [
            (key, value),
            ("doctop", obj["doctop"] + diff),
            ("height", obj["height"] - diff),
        ]
        if "y1" in obj:
            new_items += [
                ("y1", obj["y1"] - diff),
            ]
    if key == "bottom":
        assert(value >= obj["top"])
        new_items = [
            (key, value),
            ("height", obj["height"] + diff),
        ]
        if "y0" in obj:
            new_items += [
                ("y0", obj["y0"] - diff),
            ]
    return obj.__class__(tuple(obj.items()) + tuple(new_items))

def curve_to_edges(curve):
    point_pairs = zip(curve["points"], curve["points"][1:]) 
    return [ {
        "x0": min(p0[0], p1[0]),
        "x1": max(p0[0], p1[0]),
        "top": min(p0[1], p1[1]),
        "doctop": min(p0[1], p1[1]) + (curve["doctop"] - curve["top"]),
        "bottom": max(p0[1], p1[1]),
        "width": abs(p0[0] - p1[0]),
        "height": abs(p0[1] - p1[1]),
        "orientation": "v" if p0[0] == p1[0] else ("h" if p0[1] == p1[1] else None)
    } for p0, p1 in point_pairs ]

def rect_to_edges(rect):
    top, bottom, left, right = [ dict(rect) for x in range(4) ]
    top.update({
        "object_type": "rect_edge",
        "height": decimalize(0),
        "y0": rect["y1"],
        "bottom": rect["top"],
        "orientation": "h"
    })
    bottom.update({
        "object_type": "rect_edge",
        "height": decimalize(0),
        "y1": rect["y0"],
        "top": rect["top"] + rect["height"],
        "doctop": rect["doctop"] + rect["height"],
        "orientation": "h"
    })
    left.update({
        "object_type": "rect_edge",
        "width": decimalize(0),
        "x1": rect["x0"],
        "orientation": "v"
    })
    right.update({
        "object_type": "rect_edge",
        "width": decimalize(0),
        "x0": rect["x1"],
        "orientation": "v"
    })
    return [ top, bottom, left, right ]

def line_to_edge(line):
    edge = dict(line)
    edge["orientation"] = "h" if (line["top"] == line["bottom"]) else "v"
    return edge

def obj_to_edges(obj):
    return {
        "line": lambda x: [ line_to_edge(x) ],
        "rect": rect_to_edges,
        "rect_edge": rect_to_edges,
        "curve": curve_to_edges,
    }[obj["object_type"]](obj)

def filter_edges(edges, orientation=None,
    edge_type=None,
    min_length=1):

    if orientation not in ("v", "h", None):
        raise ValueError("Orientation must be 'v' or 'h'")

    def test(e):
        dim = "height" if e["orientation"] == "v" else "width"
        et = (e["object_type"] == edge_type if edge_type != None else True)
        return et & (
            (True if orientation == None else (e["orientation"] == orientation)) & 
            (e[dim] >= min_length)
        )

    edges = filter(test, edges)
    return list(edges)



import PIL.Image
import PIL.ImageDraw
import wand.image
import sys, os
from io import BytesIO
from pdfplumber import utils
from pdfplumber.table import TableFinder

class COLORS(object):
    RED = (255, 0, 0)
    GREEN = (0, 255, 0)
    BLUE = (0, 0, 255)
    TRANSPARENT = (0, 0, 0, 0)

DEFAULT_FILL = COLORS.BLUE + (50,)
DEFAULT_STROKE = COLORS.RED + (200,)
DEFAULT_STROKE_WIDTH = 1
DEFAULT_RESOLUTION = 72

def get_page_image(stream, page_no, resolution):
    """
    For kwargs, see http://docs.wand-py.org/en/latest/wand/image.html#wand.image.Image
    """
    stream.seek(0)
    with wand.image.Image(file=stream, resolution=resolution) as pages:
        img = wand.image.Image(image=pages.sequence[page_no])
        if img.alpha_channel:
            img.background_color = wand.image.Color('white')
            img.alpha_channel = 'background'
        with img.convert("png") as png:
            im = PIL.Image.open(BytesIO(png.make_blob()))
            if "transparency" in im.info:
                converted = im.convert("RGBA").convert("RGB")
            else:
                converted = im.convert("RGB")
            return converted

class PageImage(object):
    def __init__(self, page, original=None, resolution=DEFAULT_RESOLUTION):
        self.page = page
        if original == None:
            self.original = get_page_image(
                page.pdf.stream,
                page.page_number - 1,
                resolution
            )
        else:
            self.original = original

        d = self.page.decimalize
        self.decimalize = d
        if page.is_original:
            self.root = page
            cropped = False
        else:
            self.root = page.root_page
            cropped = page.root_page.bbox != page.bbox
        self.scale = d(self.original.size[0]) / d(self.root.width)
        if cropped:
            cropbox = (
                (page.bbox[0] - page.root_page.bbox[0]) * self.scale,
                (page.bbox[1] - page.root_page.bbox[1]) * self.scale,
                (page.bbox[2] - page.root_page.bbox[0]) * self.scale,
                (page.bbox[3] - page.root_page.bbox[1]) * self.scale,
            )
            self.original = self.original.crop(map(int, cropbox))
        self.reset()

    def _reproject_bbox(self, bbox):
        x0, top, x1, bottom = bbox
        _x0, _top = self._reproject((x0, top))
        _x1, _bottom = self._reproject((x1, bottom))
        return (_x0, _top, _x1, _bottom)

    def _reproject(self, coord):
        """Given an (x0, top) tuple from the *root* coordinate system,
        return an (x0, top) tuple in the *image* coordinate system.
        """
        x0, top = coord
        px0, ptop = self.page.bbox[:2]
        rx0, rtop = self.root.bbox[:2]
        _x0 = (x0 + rx0 - px0) * self.scale
        _top = (top + rtop - ptop) * self.scale
        return (_x0, _top)

    def reset(self):
        self.annotated = PIL.Image.new(self.original.mode, self.original.size)
        self.annotated.paste(self.original)
        self.draw = PIL.ImageDraw.Draw(self.annotated, "RGBA")
        return self

    def copy(self):
        return self.__class__(self.page, self.original)

    def draw_line(self, points_or_obj,
        stroke=DEFAULT_STROKE,
        stroke_width=DEFAULT_STROKE_WIDTH):
        if isinstance(points_or_obj, (tuple, list)):
            points = points_or_obj
        elif type(points_or_obj) == dict and "points" in points_or_obj:
            points = points_or_obj["points"]
        else:
            obj = points_or_obj
            points = ((obj["x0"], obj["top"]), (obj["x1"], obj["bottom"]))
        self.draw.line(
            list(map(self._reproject, points)),
            fill=stroke,
            width=stroke_width
        )
        return self

    def draw_lines(self, list_of_lines, **kwargs):
        for x in utils.to_list(list_of_lines):
            self.draw_line(x, **kwargs)
        return self
        
    def draw_vline(self, location,
        stroke=DEFAULT_STROKE,
        stroke_width=DEFAULT_STROKE_WIDTH):
        points = (location, self.page.bbox[1], location, self.page.bbox[3])
        self.draw.line(
            self._reproject_bbox(points),
            fill=stroke,
            width=stroke_width
        )
        return self

    def draw_vlines(self, locations, **kwargs):
        for x in utils.to_list(locations):
            self.draw_vline(x, **kwargs)
        return self
        
    def draw_hline(self, location,
        stroke=DEFAULT_STROKE,
        stroke_width=DEFAULT_STROKE_WIDTH):
        points = (self.page.bbox[0], location, self.page.bbox[2], location)
        self.draw.line(
            self._reproject_bbox(points),
            fill=stroke,
            width=stroke_width
        )
        return self

    def draw_hlines(self, locations, **kwargs):
        for x in utils.to_list(locations):
            self.draw_hline(x, **kwargs)
        return self
        
    def draw_rect(self, bbox_or_obj,
        fill=DEFAULT_FILL,
        stroke=DEFAULT_STROKE,
        stroke_width=DEFAULT_STROKE_WIDTH):
        if isinstance(bbox_or_obj, (tuple, list)):
            bbox = bbox_or_obj
        else:
            obj = bbox_or_obj
            bbox = (obj["x0"], obj["top"], obj["x1"], obj["bottom"])

        x0, top, x1, bottom = bbox
        half = self.decimalize(stroke_width / 2)
        x0 += half
        top += half
        x1 -= half
        bottom -= half

        self.draw.rectangle(
            self._reproject_bbox((x0, top, x1, bottom)),
            fill,
            COLORS.TRANSPARENT
        )

        if stroke_width > 0:
            segments = [
                ((x0, top), (x1, top)), # top
                ((x0, bottom), (x1, bottom)), # bottom
                ((x0, top), (x0, bottom)), # left
                ((x1, top), (x1, bottom)), # right
            ]
            self.draw_lines(
                segments,
                stroke=stroke,
                stroke_width=stroke_width
            )
        return self

    def draw_rects(self, list_of_rects, **kwargs):
        for x in utils.to_list(list_of_rects):
            self.draw_rect(x, **kwargs)
        return self

    def draw_circle(self, center_or_obj,
        radius=5,
        fill=DEFAULT_FILL,
        stroke=DEFAULT_STROKE):
        if isinstance(center_or_obj, (tuple, list)):
            center = center_or_obj
        else:
            obj = center_or_obj
            center = (
                (obj["x0"] + obj["x1"]) / 2,
                (obj["top"] + obj["bottom"]) / 2
            )
        cx, cy = center
        bbox = self.decimalize((
            cx - radius,
            cy - radius,
            cx + radius,
            cy + radius
        ))
        self.draw.ellipse(
            self._reproject_bbox(bbox),
            fill,
            stroke
        )
        return self

    def draw_circles(self, list_of_circles, **kwargs):
        for x in utils.to_list(list_of_circles):
            self.draw_circle(x, **kwargs)
        return self

    def save(self, *args, **kwargs):
        return self.annotated.save(*args, **kwargs)

    def debug_table(self, table,
        fill=DEFAULT_FILL,
        stroke=DEFAULT_STROKE,
        stroke_width=1):
        """
        Outline all found tables.
        """
        self.draw_rects(table.cells,
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width)
        return self

    def debug_tablefinder(self, tf={}):
        if isinstance(tf, TableFinder):
            pass
        elif isinstance(tf, dict):
            tf = self.page.debug_tablefinder(tf)
        else:
            raise ValueError("Argument must be instance of TableFinder or a TableFinder settings dict.")

        for table in tf.tables:
            self.debug_table(table)

        self.draw_lines(tf.edges, stroke_width=1)

        self.draw_circles(tf.intersections.keys(),
            fill=COLORS.TRANSPARENT,
            stroke=COLORS.BLUE + (200,),
            radius=3)
        return self
        
    def outline_words(self,
        stroke=DEFAULT_STROKE,
        fill=DEFAULT_FILL,
        stroke_width=DEFAULT_STROKE_WIDTH,
        x_tolerance=utils.DEFAULT_X_TOLERANCE,
        y_tolerance=utils.DEFAULT_Y_TOLERANCE):

        words = self.page.extract_words(x_tolerance=x_tolerance, y_tolerance=y_tolerance)
        self.draw_rects(words, stroke=stroke, fill=fill, stroke_width=stroke_width)
        return self

    def outline_chars(self,
        stroke=(255, 0, 0, 255),
        fill=(255, 0, 0, int(255/4)),
        stroke_width=DEFAULT_STROKE_WIDTH):

        self.draw_rects(self.page.chars, stroke=stroke, fill=fill, stroke_width=stroke_width)
        return self

    def _repr_png_(self):
        b = BytesIO()
        self.annotated.save(b, 'PNG')
        return b.getvalue()

